/*
 * Copyright (c) 2016-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jul  6 2016
 */
#include "acdynamixel.h"

#include <math.h>

#include "dynamixel_c_types.h"


/* --- Attribute set_position ------------------------------------------- */

/** Validation codel dx_set_position_validate of attribute set_position.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
dx_set_position_validate(sequence8_double *position,
                         sequence8_double *velocity,
                         sequence8_double *effort,
                         const genom_context self)
{
  size_t l;

  /* discard trailing nans */
  l = position->_length;
  while(l > 0 && isnan(position->_buffer[l-1])) l--;
  position->_length = l;

  /* reset other input */
  velocity->_length = 0;
  effort->_length = 0;

  return genom_ok;
}


/* --- Attribute set_velocity ------------------------------------------- */

/** Validation codel dx_set_velocity_validate of attribute set_velocity.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
dx_set_velocity_validate(sequence8_double *position,
                         sequence8_double *velocity,
                         sequence8_double *effort,
                         const genom_context self)
{
  size_t l;

  /* discard trailing nans */
  l = velocity->_length;
  while(l > 0 && isnan(velocity->_buffer[l-1])) l--;
  velocity->_length = l;

  /* reset other input */
  position->_length = 0;
  effort->_length = 0;

  return genom_ok;
}


/* --- Attribute set_effort --------------------------------------------- */

/** Validation codel dx_set_effort_validate of attribute set_effort.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
dx_set_effort_validate(sequence8_double *position,
                       sequence8_double *velocity,
                       sequence8_double *effort,
                       const genom_context self)
{
  size_t l;

  /* discard trailing nans */
  l = effort->_length;
  while(l > 0 && isnan(effort->_buffer[l-1])) l--;
  effort->_length = l;

  /* reset other input */
  position->_length = 0;
  velocity->_length = 0;

  return genom_ok;
}


/* --- Attribute set_position_effort ------------------------------------ */

/** Validation codel dx_set_position_effort_validate of attribute set_position_effort.
 *
 * Returns genom_ok.
 * Throws .
 */
genom_event
dx_set_position_effort_validate(sequence8_double *position,
                                sequence8_double *velocity,
                                sequence8_double *effort,
                                const genom_context self)
{
  size_t l;

  /* discard trailing nans */
  l = position->_length;
  while(l > 0 && isnan(position->_buffer[l-1])) l--;
  position->_length = l;

  l = effort->_length;
  while(l > 0 && isnan(effort->_buffer[l-1])) l--;
  effort->_length = l;

  /* reset other input */
  velocity->_length = 0;

  return genom_ok;
}


/* --- Function set_position_pidff -------------------------------------- */

/** Codel dx_set_position_pidff of function set_position_pidff.
 *
 * Returns genom_ok.
 * Throws dynamixel_e_range.
 */
genom_event
dx_set_position_pidff(uint16_t id, double kP, double kI, double kD,
                      double kFF1, double kFF2,
                      dynamixel_ids_update_s *update,
                      sequence8_dynamixel_pidff_s *position_pidff,
                      const genom_context self)
{
  if (id >= or_joint_max_joints) return dynamixel_e_range(self);
  if (kP < 0 || kP > 16383/128. ||
      kI < 0 || kI > 16383/65536. ||
      kD < 0 || kD > 16383/16. ||
      kFF1 < 0 || kFF1 > 16383/4. ||
      kFF2 < 0 || kFF2 > 16383/4.)
    return dynamixel_e_range(self);

  if (id >= position_pidff->_length) position_pidff->_length = id + 1;

  position_pidff->_buffer[id] = (struct dynamixel_pidff_s){
    .kP = kP, .kI = kI, .kD = kD, .kFF1 = kFF1, .kFF2 = kFF2
  };
  update->position_pidff[id] = true;
  update->update = true;
  return genom_ok;
}


/* --- Function set_velocity_pi ----------------------------------------- */

/** Codel dx_set_velocity_pi of function set_velocity_pi.
 *
 * Returns genom_ok.
 * Throws dynamixel_e_range.
 */
genom_event
dx_set_velocity_pi(uint16_t id, double kP, double kI,
                   dynamixel_ids_update_s *update,
                   sequence8_dynamixel_pi_s *velocity_pi,
                   const genom_context self)
{
  if (id >= or_joint_max_joints) return dynamixel_e_range(self);
  if (kP < 0 || kP > 16383/128. ||
      kI < 0 || kI > 16383/65536.)
    return dynamixel_e_range(self);

  if (id >= velocity_pi->_length) velocity_pi->_length = id + 1;

  velocity_pi->_buffer[id] = (struct dynamixel_pi_s){
    .kP = kP, .kI = kI
  };
  update->velocity_pi[id] = true;
  update->update = true;
  return genom_ok;
}


/* --- Function set_offset ---------------------------------------------- */

/** Codel dx_set_offset of function set_offset.
 *
 * Returns genom_ok.
 * Throws dynamixel_e_range.
 */
genom_event
dx_set_offset(uint16_t id, double position,
              dynamixel_ids_update_s *update, sequence8_double *offset,
              const genom_context self)
{
  if (id >= or_joint_max_joints) return dynamixel_e_range(self);

  if (id >= offset->_length) offset->_length = id + 1;
  offset->_buffer[id] = position;

  update->offset[id] = true;
  update->update = true;
  return genom_ok;
}


/* --- Function set_profile --------------------------------------------- */

/** Codel dx_set_profile of function set_profile.
 *
 * Returns genom_ok.
 * Throws dynamixel_e_range.
 */
genom_event
dx_set_profile(uint16_t id, const dynamixel_profile_s *desired,
               const sequence8_dynamixel_limits_s *limits,
               dynamixel_ids_update_s *update,
               sequence8_dynamixel_profile_s *profile,
               const genom_context self)
{
  if (id >= or_joint_max_joints) return dynamixel_e_range(self);
  if (desired->velocity < 0. || desired->acceleration < 0. ||
      desired->velocity > limits->_buffer[id].velocity_max ||
      desired->acceleration > limits->_buffer[id].acceleration_max)
    return dynamixel_e_range(self);

  if (id >= profile->_length) profile->_length = id + 1;
  profile->_buffer[id] = *desired;

  update->profile[id] = true;
  update->update = true;
  return genom_ok;
}
