/*
 * Copyright (c) 2015-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Wed Jul  6 2016
 */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include "codels.h"


static uint16_t	crc16_8005(uint16_t crc, uint8_t *data, size_t size);


/* --- dx_open_tty --------------------------------------------------------- */

/* Open a serial port, configure to the given baud rate */
int
dx_open_tty(const char *device, uint32_t speed)
{
  struct termios t;
  speed_t baud;
  int fd;

  /* select baud rate */
#ifndef B9600
# define B9600 9600U
#endif
#ifndef B57600
# define B57600 57600U
#endif
#ifndef B115200
# define B115200 115200U
#endif
#ifndef B500000
# define B500000 500000U
#endif
#ifndef B1000000
# define B1000000 1000000U
#endif
#ifndef B2000000
# define B2000000 2000000U
#endif
#ifndef B3000000
# define B3000000 3000000U
#endif
  switch(speed) {
    case 9600:		baud = B9600; break;
    case 57600:		baud = B57600; break;
    case 115200:	baud = B115200; break;
    case 1000000:	baud = B1000000; break;
    case 2000000:	baud = B2000000; break;
    case 3000000:	baud = B3000000; break;

    default: errno = EINVAL; return -1;
  }

  /* open non-blocking */
  fd = open(device, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd < 0) return fd;
  if (!isatty(fd)) {
    errno = ENOTTY;
    return -1;
  }

  /* configure line discipline */
  if (tcgetattr(fd, &t)) return -1;

  t.c_iflag = IGNBRK;
  t.c_oflag = 0;
  t.c_lflag = 0;
  t.c_cflag = CS8 | CREAD | CLOCAL;
  t.c_cc[VMIN] = 0;
  t.c_cc[VTIME] = 0;

  if (cfsetospeed(&t, baud)) return -1;
  if (cfsetispeed(&t, baud)) return -1;

  if (tcsetattr(fd, TCSANOW, &t)) return -1;

  /* discard any pending data */
  tcflush(fd, TCIOFLUSH);

  return fd;
}


/* --- dx_wait_msg --------------------------------------------------------- */

int
dx_wait_msg(const struct dynamixel_channel_s *channel, int timeout)
{
  struct pollfd pfd;
  int s;

  pfd.fd = channel->fd;
  pfd.events = POLLIN;

  do {
    s = poll(&pfd, pfd.fd < 0 ? 0 : 1, timeout);
  } while(s < 0 && errno == EINTR);
  if (s > 0 && pfd.revents & POLLHUP) return -2;

  return s;
}


/* --- dx_recv_msg --------------------------------------------------------- */

/* returns: 0: timeout/incomplete, -1: error, 1: complete msg */

int
dx_recv_msg(struct dynamixel_channel_s *chan, bool block)
{
  struct iovec iov[2];
  ssize_t s;
  uint8_t c;

  if (chan->fd < 0) return -1;

  do {
    /* feed the ring  buffer */
    iov[0].iov_base = chan->buf + chan->w;
    iov[1].iov_base = chan->buf;

    /* at most n-1 of the n available bytes can be used */
    if (chan->r > chan->w) {
      iov[0].iov_len = chan->r - chan->w - 1;
      iov[1].iov_len = 0;
    } else if (chan->r > 0) {
      iov[0].iov_len = sizeof(chan->buf) - chan->w;
      iov[1].iov_len = chan->r - 1;
    } else {
      iov[0].iov_len = sizeof(chan->buf) - chan->w - 1;
      iov[1].iov_len = 0;
    }

    if (iov[0].iov_len || iov[1].iov_len) {
      do {
        s = readv(chan->fd, iov, 2);
      } while(s < 0 && errno == EINTR);

      if (s < 0) return -1;
      else if (s == 0 && (chan->state || chan->r == chan->w)) {
        if (chan->state || block) {
          struct pollfd fd = { .fd = chan->fd, .events = POLLIN };

          s = poll(&fd, 1, 10/*ms*/);
          if (s >= 0 && fd.revents & POLLHUP) return -2;
          if (s) continue;
        }
        return 0;
      } else
        chan->w = (chan->w + s) % sizeof(chan->buf);
    }

    while(chan->r != chan->w) {
      c = chan->buf[chan->r];
      chan->r = (chan->r + 1) % sizeof(chan->buf);

      chan->header = (chan->header << 8) | c;
      if (chan->header == 0xfffffd00) {
        if (chan->state) warnx("dropped partial message");
        chan->state = 1; continue;
      }

      switch(chan->state) {
        case 0: break;

          /* id */
        case 1:
          chan->id = c;
          chan->sum = crc16_8005(0, (uint8_t []){0xff, 0xff, 0xfd, 0, c}, 5);
          chan->state++; break;

          /* len */
        case 2:
          chan->len = c;
          chan->sum = crc16_8005(chan->sum, &c, 1);
          chan->state++; break;
        case 3:
          chan->len += c << 8;
          if (chan->len < 4) {
            warnx("dropped message with bad length");
            chan->state = 0; break;
          }
          chan->len -= 4;
          chan->p = 0;
          chan->sum = crc16_8005(chan->sum, &c, 1);
          chan->state++; break;

          /* status */
        case 4:
          if (c != 0x55) {
            warnx("dropped unknown status message 0x%02x", c);
            chan->state = 0; break;
          }
          chan->sum = crc16_8005(chan->sum, &c, 1);
          chan->state++; break;

          /* error */
        case 5:
          chan->error = c;
          chan->sum = crc16_8005(chan->sum, &c, 1);
          chan->state++; break;

          /* data */
        case 6:
          if (!chan->len) { chan->state = 0; break; }

          if (chan->header != 0xfffffdfd) {
            /* dynamixel "byte stuffing": ignore the last 0xfd in data when
             * 0xfffffdfd is seen, but include it in CRC computation */
            chan->param[chan->p++] = c;
          }
          chan->sum = crc16_8005(chan->sum, &c, 1);
          chan->len--;
          if (!chan->len) { chan->state++; }
          break;

          /* crc */
        case 7:
          if (c != (chan->sum & 0xff)) {
            warnx("dropped message with bad checksum");
            chan->state = 0; break;
          }
          chan->state++; break;
        case 8:
          chan->state = 0;
          if (c != (chan->sum >> 8)) {
            warnx("dropped message with bad checksum");
            break;
          }
          return 1;
      }
    }
  } while(1);

  return 0;
}


/* --- dx_send_msg --------------------------------------------------------- */

int
dx_send_msg(const struct dynamixel_channel_s *chan,
            uint8_t id, enum dx_cmd cmd, ...)
{
  va_list ap;
  ssize_t s;
  uint8_t buf[64], *w, *r;
  uint16_t cksum;

  if (chan->fd < 0) return -1;

  /* encode message */
  va_start(ap, cmd);
  w = buf;
  *w++ = 0xff;
  *w++ = 0xff;
  *w++ = 0xfd;
  *w++ = 0;

  switch(cmd) {
    case DX_PING:
      *w++ = id;
      *w++ = 3; *w++ = 0;
      *w++ = cmd;
      break;

    case DX_READ: {
      uint16_t n = va_arg(ap, int);
      uint16_t addr = va_arg(ap, int);

      *w++ = id;
      *w++ = 7; *w++ = 0;
      *w++ = cmd;
      *w++ = addr & 0xff; *w++ = addr >> 8;
      *w++ = n & 0xff; *w++ = n >> 8;
      break;
    }
    case DX_SYNC_READ: {
      uint16_t n = va_arg(ap, int);
      uint16_t addr = va_arg(ap, int);
      int i;

      *w++ = 0xfe;
      *w++ = 7 + id; *w++ = 0;
      *w++ = cmd;
      *w++ = addr & 0xff; *w++ = addr >> 8;
      *w++ = n & 0xff; *w++ = n >> 8;
      for(i = 1; i <= id; i++) *w++ = i;
      break;
    }

    case DX_WRITE: {
      int n = va_arg(ap, int);
      uint16_t addr = va_arg(ap, int);
      uint8_t d;
      int i;

      *w++ = id;
      *w++ = 5 + n; *w++ = 0;
      *w++ = cmd;
      *w++ = addr & 0xff; *w++ = addr >> 8;
      for(i = 0; i < n; i++) {
        d = va_arg(ap, int);
        *w++ = d;
      }
      break;
    }

    case DX_SYNC_WRITE: {
      uint32_t stuff; /* for dynamixel "byte stuffing" */
      uint16_t n = va_arg(ap, int);
      uint16_t addr = va_arg(ap, int);
      uint8_t *d = va_arg(ap, uint8_t *);
      uint8_t *len;
      int i, j;

      *w++ = 0xfe;
      len = w; *w++ = 7 + (n + 1) * id; *w++ = 0;
      *w++ = cmd;
      *w++ = addr & 0xff; *w++ = addr >> 8;
      *w++ = n & 0xff; *w++ = n >> 8;
      stuff = 0;
      for(i = 1; i <= id; i++) {
        *w++ = i;
        for(j = 0; j < n; j++) {
          stuff = ((stuff << 8) & 0xffffff) | *d;
          *w++ = *d++;
          if (stuff == 0xfffffd) {
            /* "byte stuffing", add extra 0xfd */
            *w++ = 0xfd;
            (*len)++;
          }
        }
      }
      break;
    }
  }

  cksum = crc16_8005(0, buf, w-buf);
  *w++ = cksum & 0xff; *w++ = cksum >> 8;
  va_end(ap);
  assert(w - buf < sizeof(buf));

  /* send */
  r = buf;
  while (w > r) {
    do {
      s = write(chan->fd, r, w - r);
    } while (s < 0 && errno == EINTR);
    if (s < 0) return -1;

    r += s;
  }

  return 0;
}


/* --- crc16_8005 ---------------------------------------------------------- */

/** Compute a CRC16 cksum with polynomial x^16 + x^15 + x^2 + 1 */

static uint16_t
crc16_8005(uint16_t crc, uint8_t *data, size_t size)
{
  static const uint16_t crc_table[256] = {
    0x0000, 0x8005, 0x800F, 0x000A, 0x801B, 0x001E, 0x0014, 0x8011,
    0x8033, 0x0036, 0x003C, 0x8039, 0x0028, 0x802D, 0x8027, 0x0022,
    0x8063, 0x0066, 0x006C, 0x8069, 0x0078, 0x807D, 0x8077, 0x0072,
    0x0050, 0x8055, 0x805F, 0x005A, 0x804B, 0x004E, 0x0044, 0x8041,
    0x80C3, 0x00C6, 0x00CC, 0x80C9, 0x00D8, 0x80DD, 0x80D7, 0x00D2,
    0x00F0, 0x80F5, 0x80FF, 0x00FA, 0x80EB, 0x00EE, 0x00E4, 0x80E1,
    0x00A0, 0x80A5, 0x80AF, 0x00AA, 0x80BB, 0x00BE, 0x00B4, 0x80B1,
    0x8093, 0x0096, 0x009C, 0x8099, 0x0088, 0x808D, 0x8087, 0x0082,
    0x8183, 0x0186, 0x018C, 0x8189, 0x0198, 0x819D, 0x8197, 0x0192,
    0x01B0, 0x81B5, 0x81BF, 0x01BA, 0x81AB, 0x01AE, 0x01A4, 0x81A1,
    0x01E0, 0x81E5, 0x81EF, 0x01EA, 0x81FB, 0x01FE, 0x01F4, 0x81F1,
    0x81D3, 0x01D6, 0x01DC, 0x81D9, 0x01C8, 0x81CD, 0x81C7, 0x01C2,
    0x0140, 0x8145, 0x814F, 0x014A, 0x815B, 0x015E, 0x0154, 0x8151,
    0x8173, 0x0176, 0x017C, 0x8179, 0x0168, 0x816D, 0x8167, 0x0162,
    0x8123, 0x0126, 0x012C, 0x8129, 0x0138, 0x813D, 0x8137, 0x0132,
    0x0110, 0x8115, 0x811F, 0x011A, 0x810B, 0x010E, 0x0104, 0x8101,
    0x8303, 0x0306, 0x030C, 0x8309, 0x0318, 0x831D, 0x8317, 0x0312,
    0x0330, 0x8335, 0x833F, 0x033A, 0x832B, 0x032E, 0x0324, 0x8321,
    0x0360, 0x8365, 0x836F, 0x036A, 0x837B, 0x037E, 0x0374, 0x8371,
    0x8353, 0x0356, 0x035C, 0x8359, 0x0348, 0x834D, 0x8347, 0x0342,
    0x03C0, 0x83C5, 0x83CF, 0x03CA, 0x83DB, 0x03DE, 0x03D4, 0x83D1,
    0x83F3, 0x03F6, 0x03FC, 0x83F9, 0x03E8, 0x83ED, 0x83E7, 0x03E2,
    0x83A3, 0x03A6, 0x03AC, 0x83A9, 0x03B8, 0x83BD, 0x83B7, 0x03B2,
    0x0390, 0x8395, 0x839F, 0x039A, 0x838B, 0x038E, 0x0384, 0x8381,
    0x0280, 0x8285, 0x828F, 0x028A, 0x829B, 0x029E, 0x0294, 0x8291,
    0x82B3, 0x02B6, 0x02BC, 0x82B9, 0x02A8, 0x82AD, 0x82A7, 0x02A2,
    0x82E3, 0x02E6, 0x02EC, 0x82E9, 0x02F8, 0x82FD, 0x82F7, 0x02F2,
    0x02D0, 0x82D5, 0x82DF, 0x02DA, 0x82CB, 0x02CE, 0x02C4, 0x82C1,
    0x8243, 0x0246, 0x024C, 0x8249, 0x0258, 0x825D, 0x8257, 0x0252,
    0x0270, 0x8275, 0x827F, 0x027A, 0x826B, 0x026E, 0x0264, 0x8261,
    0x0220, 0x8225, 0x822F, 0x022A, 0x823B, 0x023E, 0x0234, 0x8231,
    0x8213, 0x0216, 0x021C, 0x8219, 0x0208, 0x820D, 0x8207, 0x0202
  };

  for(; size; data++, size--)
    crc = (crc << 8) ^ crc_table[(crc >> 8) ^ *data];

  return crc;
}
